@@ -7,25 +7,23 @@ class TestThreadedTaskDispatcher(unittest.TestCase):
         from waitress.task import ThreadedTaskDispatcher
         return ThreadedTaskDispatcher()
 
-    def test_handler_thread_task_is_None(self):
-        inst = self._makeOne()
-        inst.threads[0] = True
-        inst.queue.put(None)
-        inst.handler_thread(0)
-        self.assertEqual(inst.stop_count, -1)
-        self.assertEqual(inst.threads, {})
-
     def test_handler_thread_task_raises(self):
-        from waitress.task import JustTesting
         inst = self._makeOne()
-        inst.threads[0] = True
+        inst.threads.add(0)
         inst.logger = DummyLogger()
-        task = DummyTask(JustTesting)
+        class BadDummyTask(DummyTask):
+            def service(self):
+                super(BadDummyTask, self).service()
+                inst.stop_count += 1
+                raise Exception
+        task = BadDummyTask()
         inst.logger = DummyLogger()
-        inst.queue.put(task)
+        inst.queue.append(task)
+        inst.active_count += 1
         inst.handler_thread(0)
-        self.assertEqual(inst.stop_count, -1)
-        self.assertEqual(inst.threads, {})
+        self.assertEqual(inst.stop_count, 0)
+        self.assertEqual(inst.active_count, 0)
+        self.assertEqual(inst.threads, set())
         self.assertEqual(len(inst.logger.logged), 1)
 
     def test_set_thread_count_increase(self):
@@ -38,49 +36,54 @@ class TestThreadedTaskDispatcher(unittest.TestCase):
     def test_set_thread_count_increase_with_existing(self):
         inst = self._makeOne()
         L = []
-        inst.threads = {0: 1}
+        inst.threads = {0}
         inst.start_new_thread = lambda *x: L.append(x)
         inst.set_thread_count(2)
         self.assertEqual(L, [(inst.handler_thread, (1,))])
 
     def test_set_thread_count_decrease(self):
         inst = self._makeOne()
-        inst.threads = {'a': 1, 'b': 2}
+        inst.threads = {0, 1}
         inst.set_thread_count(1)
-        self.assertEqual(inst.queue.qsize(), 1)
-        self.assertEqual(inst.queue.get(), None)
+        self.assertEqual(inst.stop_count, 1)
 
     def test_set_thread_count_same(self):
         inst = self._makeOne()
         L = []
         inst.start_new_thread = lambda *x: L.append(x)
-        inst.threads = {0: 1}
+        inst.threads = {0}
         inst.set_thread_count(1)
         self.assertEqual(L, [])
 
-    def test_add_task(self):
+    def test_add_task_with_idle_threads(self):
         task = DummyTask()
         inst = self._makeOne()
+        inst.threads.add(0)
+        inst.queue_logger = DummyLogger()
         inst.add_task(task)
-        self.assertEqual(inst.queue.qsize(), 1)
-        self.assertTrue(task.deferred)
+        self.assertEqual(len(inst.queue), 1)
+        self.assertEqual(len(inst.queue_logger.logged), 0)
 
-    def test_add_task_defer_raises(self):
-        task = DummyTask(ValueError)
+    def test_add_task_with_all_busy_threads(self):
+        task = DummyTask()
         inst = self._makeOne()
-        self.assertRaises(ValueError, inst.add_task, task)
-        self.assertEqual(inst.queue.qsize(), 0)
-        self.assertTrue(task.deferred)
-        self.assertTrue(task.cancelled)
+        inst.queue_logger = DummyLogger()
+        inst.add_task(task)
+        self.assertEqual(len(inst.queue_logger.logged), 1)
+        inst.add_task(task)
+        self.assertEqual(len(inst.queue_logger.logged), 2)
 
     def test_shutdown_one_thread(self):
         inst = self._makeOne()
-        inst.threads[0] = 1
+        inst.threads.add(0)
         inst.logger = DummyLogger()
         task = DummyTask()
-        inst.queue.put(task)
+        inst.queue.append(task)
         self.assertEqual(inst.shutdown(timeout=.01), True)
-        self.assertEqual(inst.logger.logged, ['1 thread(s) still running'])
+        self.assertEqual(inst.logger.logged, [
+            '1 thread(s) still running',
+            'Canceling 1 pending task(s)',
+        ])
         self.assertEqual(task.cancelled, True)
 
     def test_shutdown_no_threads(self):
@@ -108,15 +111,6 @@ class TestTask(unittest.TestCase):
         inst = self._makeOne(request=request)
         self.assertEqual(inst.version, '1.0')
 
-    def test_cancel(self):
-        inst = self._makeOne()
-        inst.cancel()
-        self.assertTrue(inst.close_on_finish)
-
-    def test_defer(self):
-        inst = self._makeOne()
-        self.assertEqual(inst.defer(), None)
-
     def test_build_response_header_bad_http_version(self):
         inst = self._makeOne()
         inst.request = DummyParser()
@@ -202,6 +196,57 @@ class TestTask(unittest.TestCase):
         self.assertEqual(inst.close_on_finish, True)
         self.assertTrue(('Connection', 'close') in inst.response_headers)
 
+    def test_build_response_header_v11_204_no_content_length_or_transfer_encoding(self):
+        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
+        # for any response with a status code of 1xx or 204.
+        inst = self._makeOne()
+        inst.request = DummyParser()
+        inst.version = '1.1'
+        inst.status = '204 No Content'
+        result = inst.build_response_header()
+        lines = filter_lines(result)
+        self.assertEqual(len(lines), 4)
+        self.assertEqual(lines[0], b'HTTP/1.1 204 No Content')
+        self.assertEqual(lines[1], b'Connection: close')
+        self.assertTrue(lines[2].startswith(b'Date:'))
+        self.assertEqual(lines[3], b'Server: waitress')
+        self.assertEqual(inst.close_on_finish, True)
+        self.assertTrue(('Connection', 'close') in inst.response_headers)
+
+    def test_build_response_header_v11_1xx_no_content_length_or_transfer_encoding(self):
+        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
+        # for any response with a status code of 1xx or 204.
+        inst = self._makeOne()
+        inst.request = DummyParser()
+        inst.version = '1.1'
+        inst.status = '100 Continue'
+        result = inst.build_response_header()
+        lines = filter_lines(result)
+        self.assertEqual(len(lines), 4)
+        self.assertEqual(lines[0], b'HTTP/1.1 100 Continue')
+        self.assertEqual(lines[1], b'Connection: close')
+        self.assertTrue(lines[2].startswith(b'Date:'))
+        self.assertEqual(lines[3], b'Server: waitress')
+        self.assertEqual(inst.close_on_finish, True)
+        self.assertTrue(('Connection', 'close') in inst.response_headers)
+
+    def test_build_response_header_v11_304_no_content_length_or_transfer_encoding(self):
+        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
+        # for any response with a status code of 1xx, 204 or 304.
+        inst = self._makeOne()
+        inst.request = DummyParser()
+        inst.version = '1.1'
+        inst.status = '304 Not Modified'
+        result = inst.build_response_header()
+        lines = filter_lines(result)
+        self.assertEqual(len(lines), 4)
+        self.assertEqual(lines[0], b'HTTP/1.1 304 Not Modified')
+        self.assertEqual(lines[1], b'Connection: close')
+        self.assertTrue(lines[2].startswith(b'Date:'))
+        self.assertEqual(lines[3], b'Server: waitress')
+        self.assertEqual(inst.close_on_finish, True)
+        self.assertTrue(('Connection', 'close') in inst.response_headers)
+
     def test_build_response_header_via_added(self):
         inst = self._makeOne()
         inst.request = DummyParser()
@@ -248,6 +293,12 @@ class TestTask(unittest.TestCase):
         inst.remove_content_length_header()
         self.assertEqual(inst.response_headers, [])
 
+    def test_remove_content_length_header_with_other(self):
+        inst = self._makeOne()
+        inst.response_headers = [('Content-Length', '70'), ('Content-Type', 'text/html')]
+        inst.remove_content_length_header()
+        self.assertEqual(inst.response_headers, [('Content-Type', 'text/html')])
+
     def test_start(self):
         inst = self._makeOne()
         inst.start()
@@ -518,6 +569,34 @@ class TestWSGITask(unittest.TestCase):
         self.assertEqual(inst.close_on_finish, True)
         self.assertEqual(len(inst.logger.logged), 0)
 
+    def test_execute_app_without_body_204_logged(self):
+        def app(environ, start_response):
+            start_response('204 No Content', [('Content-Length', '3')])
+            return [b'abc']
+        inst = self._makeOne()
+        inst.channel.server.application = app
+        inst.logger = DummyLogger()
+        inst.execute()
+        self.assertEqual(inst.close_on_finish, True)
+        self.assertNotIn(b'abc', inst.channel.written)
+        self.assertNotIn(b'Content-Length', inst.channel.written)
+        self.assertNotIn(b'Transfer-Encoding', inst.channel.written)
+        self.assertEqual(len(inst.logger.logged), 1)
+
+    def test_execute_app_without_body_304_logged(self):
+        def app(environ, start_response):
+            start_response('304 Not Modified', [('Content-Length', '3')])
+            return [b'abc']
+        inst = self._makeOne()
+        inst.channel.server.application = app
+        inst.logger = DummyLogger()
+        inst.execute()
+        self.assertEqual(inst.close_on_finish, True)
+        self.assertNotIn(b'abc', inst.channel.written)
+        self.assertNotIn(b'Content-Length', inst.channel.written)
+        self.assertNotIn(b'Transfer-Encoding', inst.channel.written)
+        self.assertEqual(len(inst.logger.logged), 1)
+
     def test_execute_app_returns_closeable(self):
         class closeable(list):
             def close(self):
@@ -658,11 +737,13 @@ class TestWSGITask(unittest.TestCase):
         # nail the keys of environ
         self.assertEqual(sorted(environ.keys()), [
             'CONTENT_LENGTH', 'CONTENT_TYPE', 'HTTP_CONNECTION', 'HTTP_X_FOO',
-            'PATH_INFO', 'QUERY_STRING', 'REMOTE_ADDR', 'REQUEST_METHOD',
-            'SCRIPT_NAME', 'SERVER_NAME', 'SERVER_PORT', 'SERVER_PROTOCOL',
-            'SERVER_SOFTWARE', 'wsgi.errors', 'wsgi.file_wrapper', 'wsgi.input',
+            'PATH_INFO', 'QUERY_STRING', 'REMOTE_ADDR', 'REMOTE_HOST',
+            'REMOTE_PORT', 'REQUEST_METHOD', 'SCRIPT_NAME', 'SERVER_NAME',
+            'SERVER_PORT', 'SERVER_PROTOCOL', 'SERVER_SOFTWARE', 'wsgi.errors',
+            'wsgi.file_wrapper', 'wsgi.input', 'wsgi.input_terminated',
             'wsgi.multiprocess', 'wsgi.multithread', 'wsgi.run_once',
-            'wsgi.url_scheme', 'wsgi.version'])
+            'wsgi.url_scheme', 'wsgi.version'
+        ])
 
         self.assertEqual(environ['REQUEST_METHOD'], 'GET')
         self.assertEqual(environ['SERVER_PORT'], '80')
@@ -674,6 +755,8 @@ class TestWSGITask(unittest.TestCase):
         self.assertEqual(environ['PATH_INFO'], '/')
         self.assertEqual(environ['QUERY_STRING'], 'abc')
         self.assertEqual(environ['REMOTE_ADDR'], '127.0.0.1')
+        self.assertEqual(environ['REMOTE_HOST'], '127.0.0.1')
+        self.assertEqual(environ['REMOTE_PORT'], '39830')
         self.assertEqual(environ['CONTENT_TYPE'], 'abc')
         self.assertEqual(environ['CONTENT_LENGTH'], '10')
         self.assertEqual(environ['HTTP_X_FOO'], 'BAR')
@@ -684,6 +767,7 @@ class TestWSGITask(unittest.TestCase):
         self.assertEqual(environ['wsgi.multiprocess'], False)
         self.assertEqual(environ['wsgi.run_once'], False)
         self.assertEqual(environ['wsgi.input'], 'stream')
+        self.assertEqual(environ['wsgi.input_terminated'], True)
         self.assertEqual(inst.environ, environ)
 
     def test_get_environment_values_w_scheme_override_untrusted(self):
@@ -704,8 +788,9 @@ class TestWSGITask(unittest.TestCase):
     def test_get_environment_values_w_scheme_override_trusted(self):
         import sys
         inst = self._makeOne()
-        inst.channel.addr = ['192.168.1.1']
+        inst.channel.addr = ['192.168.1.1', 8080]
         inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-proto'}
         request = DummyParser()
         request.headers = {
             'CONTENT_TYPE': 'abc',
@@ -721,14 +806,16 @@ class TestWSGITask(unittest.TestCase):
         # nail the keys of environ
         self.assertEqual(sorted(environ.keys()), [
             'CONTENT_LENGTH', 'CONTENT_TYPE', 'HTTP_CONNECTION', 'HTTP_X_FOO',
-            'PATH_INFO', 'QUERY_STRING', 'REMOTE_ADDR', 'REQUEST_METHOD',
-            'SCRIPT_NAME', 'SERVER_NAME', 'SERVER_PORT', 'SERVER_PROTOCOL',
-            'SERVER_SOFTWARE', 'wsgi.errors', 'wsgi.file_wrapper', 'wsgi.input',
-            'wsgi.multiprocess', 'wsgi.multithread', 'wsgi.run_once',
-            'wsgi.url_scheme', 'wsgi.version'])
+            'HTTP_X_FORWARDED_PROTO', 'PATH_INFO', 'QUERY_STRING',
+            'REMOTE_ADDR', 'REMOTE_HOST', 'REQUEST_METHOD', 'SCRIPT_NAME',
+            'SERVER_NAME', 'SERVER_PORT', 'SERVER_PROTOCOL', 'SERVER_SOFTWARE',
+            'wsgi.errors', 'wsgi.file_wrapper', 'wsgi.input',
+            'wsgi.input_terminated', 'wsgi.multiprocess', 'wsgi.multithread',
+            'wsgi.run_once', 'wsgi.url_scheme', 'wsgi.version'
+            ])
 
         self.assertEqual(environ['REQUEST_METHOD'], 'GET')
-        self.assertEqual(environ['SERVER_PORT'], '80')
+        self.assertEqual(environ['SERVER_PORT'], '443')
         self.assertEqual(environ['SERVER_NAME'], 'localhost')
         self.assertEqual(environ['SERVER_SOFTWARE'], 'waitress')
         self.assertEqual(environ['SERVER_PROTOCOL'], 'HTTP/1.0')
@@ -747,12 +834,14 @@ class TestWSGITask(unittest.TestCase):
         self.assertEqual(environ['wsgi.multiprocess'], False)
         self.assertEqual(environ['wsgi.run_once'], False)
         self.assertEqual(environ['wsgi.input'], 'stream')
+        self.assertEqual(environ['wsgi.input_terminated'], True)
         self.assertEqual(inst.environ, environ)
 
     def test_get_environment_values_w_bogus_scheme_override(self):
         inst = self._makeOne()
-        inst.channel.addr = ['192.168.1.1']
+        inst.channel.addr = ['192.168.1.1', 80]
         inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-proto'}
         request = DummyParser()
         request.headers = {
             'CONTENT_TYPE': 'abc',
@@ -765,6 +854,520 @@ class TestWSGITask(unittest.TestCase):
         inst.request = request
         self.assertRaises(ValueError, inst.get_environment)
 
+    def test_get_environment_warning_other_proxy_headers(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        inst.request.headers = {
+            'X_FORWARDED_FOR': '[2001:db8::1]',
+            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https'
+        }
+        inst.channel.addr = ['192.168.1.1', 80]
+        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_count = 1
+        inst.channel.server.adj.trusted_proxy_headers = {'forwarded'}
+        inst.channel.server.adj.log_untrusted_proxy_headers = True
+        environ = inst.get_environment()
+
+        self.assertEqual(len(inst.logger.logged), 1)
+        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_get_environment_contains_all_headers_including_untrusted(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        inst.request.headers = {
+            'X_FORWARDED_FOR': '198.51.100.2',
+            'X_FORWARDED_BY': 'Waitress',
+            'X_FORWARDED_PROTO': 'https',
+            'X_FORWARDED_HOST': 'example.org',
+        }
+        headers_orig = inst.request.headers.copy()
+        inst.channel.addr = ['192.168.1.1', 80]
+        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_count = 1
+        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
+        inst.channel.server.adj.clear_untrusted_proxy_headers = False
+        environ = inst.get_environment()
+
+        for k, expected in headers_orig.items():
+            result = environ['HTTP_%s' % k]
+            self.assertEqual(result, expected)
+
+    def test_get_environment_contains_only_trusted_headers(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        inst.request.headers = {
+            'X_FORWARDED_FOR': '198.51.100.2',
+            'X_FORWARDED_BY': 'Waitress',
+            'X_FORWARDED_PROTO': 'https',
+            'X_FORWARDED_HOST': 'example.org',
+        }
+        inst.channel.addr = ['192.168.1.1', 80]
+        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_count = 1
+        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
+        inst.channel.server.adj.clear_untrusted_proxy_headers = True
+        environ = inst.get_environment()
+
+        self.assertEqual(environ['HTTP_X_FORWARDED_BY'], 'Waitress')
+        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)
+        self.assertNotIn('HTTP_X_FORWARDED_PROTO', environ)
+        self.assertNotIn('HTTP_X_FORWARDED_HOST', environ)
+
+    def test_get_environment_clears_headers_if_untrusted_proxy(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        inst.request.headers = {
+            'X_FORWARDED_FOR': '198.51.100.2',
+            'X_FORWARDED_BY': 'Waitress',
+            'X_FORWARDED_PROTO': 'https',
+            'X_FORWARDED_HOST': 'example.org',
+        }
+        inst.channel.addr = ['192.168.1.255', 80]
+        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
+        inst.channel.server.adj.trusted_proxy_count = 1
+        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
+        inst.channel.server.adj.clear_untrusted_proxy_headers = True
+        environ = inst.get_environment()
+
+        self.assertNotIn('HTTP_X_FORWARDED_BY', environ)
+        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)
+        self.assertNotIn('HTTP_X_FORWARDED_PROTO', environ)
+        self.assertNotIn('HTTP_X_FORWARDED_HOST', environ)
+
+    def test_parse_proxy_headers_forwarded_for(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-for'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '192.0.2.1')
+
+    def test_parse_proxy_headers_forwarded_for_v6_missing_brackets(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '2001:db8::0'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-for'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::0')
+
+    def test_parse_proxy_headers_forwared_for_multiple(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'x-forwarded-for'}
+        )
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+
+    def test_parse_forwarded_multiple_proxies_trust_only_two(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For=192.0.2.1;host=fake.com, For=198.51.100.2;host=example.com:8080, For=203.0.113.1'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+
+    def test_parse_forwarded_multiple_proxies(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'for="[2001:db8::1]:3821";host="example.com:8443";proto="https", for=192.0.2.1;host="example.internal:8080"'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::1')
+        self.assertEqual(environ['REMOTE_PORT'], '3821')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8443')
+        self.assertEqual(environ['SERVER_PORT'], '8443')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_parse_forwarded_multiple_proxies_minimal(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'for="[2001:db8::1]";proto="https", for=192.0.2.1;host="example.org"'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::1')
+        self.assertEqual(environ['SERVER_NAME'], 'example.org')
+        self.assertEqual(environ['HTTP_HOST'], 'example.org')
+        self.assertEqual(environ['SERVER_PORT'], '443')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_parse_proxy_headers_forwarded_host_with_port(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com:8080',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+
+    def test_parse_proxy_headers_forwarded_host_without_port(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com')
+        self.assertEqual(environ['SERVER_PORT'], '80')
+
+    def test_parse_proxy_headers_forwarded_host_with_forwarded_port(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com',
+            'X_FORWARDED_PORT': '8080'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+
+    def test_parse_proxy_headers_forwarded_host_multiple_with_forwarded_port(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com, example.org',
+            'X_FORWARDED_PORT': '8080'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=2,
+            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+
+    def test_parse_proxy_headers_forwarded_host_multiple_with_forwarded_port_limit_one_trusted(self):
+        inst = self._makeOne()
+
+        headers = {
+            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com, example.org',
+            'X_FORWARDED_PORT': '8080'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '203.0.113.1')
+        self.assertEqual(environ['SERVER_NAME'], 'example.org')
+        self.assertEqual(environ['HTTP_HOST'], 'example.org:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+
+    def test_parse_forwarded(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For=198.51.100.2:5858;host=example.com:8080;proto=https'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['REMOTE_PORT'], '5858')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_parse_forwarded_empty_pair(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For=198.51.100.2;;proto=https;by=_unused'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+
+    def test_parse_forwarded_pair_token_whitespace(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For=198.51.100.2; proto =https'
+        }
+        environ = {}
+
+        with self.assertRaises(ValueError):
+            inst.parse_proxy_headers(
+                environ,
+                headers,
+                trusted_proxy_count=1,
+                trusted_proxy_headers={'forwarded'}
+            )
+
+    def test_parse_forwarded_pair_value_whitespace(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For= "198.51.100.2"; proto =https'
+        }
+        environ = {}
+
+        with self.assertRaises(ValueError):
+            inst.parse_proxy_headers(
+                environ,
+                headers,
+                trusted_proxy_count=1,
+                trusted_proxy_headers={'forwarded'}
+            )
+
+    def test_parse_forwarded_pair_no_equals(self):
+        inst = self._makeOne()
+
+        headers = {
+            'FORWARDED': 'For'
+        }
+        environ = {}
+
+        with self.assertRaises(ValueError):
+            inst.parse_proxy_headers(
+                environ,
+                headers,
+                trusted_proxy_count=1,
+                trusted_proxy_headers={'forwarded'}
+            )
+
+    def test_parse_forwarded_warning_unknown_token(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https;unknown="yolo"'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'forwarded'}
+        )
+
+        self.assertEqual(len(inst.logger.logged), 1)
+        self.assertIn('Unknown Forwarded token', inst.logger.logged[0])
+
+        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
+        self.assertEqual(environ['SERVER_PORT'], '8080')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_parse_no_valid_proxy_headers(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'X_FORWARDED_FOR': '198.51.100.2',
+            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https'
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+        )
+
+        self.assertEqual(environ, {})
+
+    def test_parse_multiple_x_forwarded_proto(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'X_FORWARDED_PROTO': 'http, https',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-proto'}
+        )
+
+        self.assertEqual(environ, {})
+        self.assertEqual(len(inst.logger.logged), 1)
+        self.assertIn("Found multiple values in X-Forwarded-Proto", inst.logger.logged[0])
+
+    def test_parse_multiple_x_forwarded_port(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'X_FORWARDED_PORT': '443, 80',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-port'}
+        )
+
+        self.assertEqual(environ, {})
+        self.assertEqual(len(inst.logger.logged), 1)
+        self.assertIn("Found multiple values in X-Forwarded-Port", inst.logger.logged[0])
+
+    def test_parse_forwarded_port_wrong_proto_port_80(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'X_FORWARDED_PORT': '80',
+            'X_FORWARDED_PROTO': 'https',
+            'X_FORWARDED_HOST': 'example.com',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-port', 'x-forwarded-host', 'x-forwarded-proto'}
+        )
+
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:80')
+        self.assertEqual(environ['SERVER_PORT'], '80')
+        self.assertEqual(environ['wsgi.url_scheme'], 'https')
+
+    def test_parse_forwarded_port_wrong_proto_port_443(self):
+        inst = self._makeOne()
+        inst.logger = DummyLogger()
+
+        headers = {
+            'X_FORWARDED_PORT': '443',
+            'X_FORWARDED_PROTO': 'http',
+            'X_FORWARDED_HOST': 'example.com',
+        }
+        environ = {}
+        inst.parse_proxy_headers(
+            environ,
+            headers,
+            trusted_proxy_count=1,
+            trusted_proxy_headers={'x-forwarded-port', 'x-forwarded-host', 'x-forwarded-proto'}
+        )
+
+        self.assertEqual(environ['SERVER_NAME'], 'example.com')
+        self.assertEqual(environ['HTTP_HOST'], 'example.com:443')
+        self.assertEqual(environ['SERVER_PORT'], '443')
+        self.assertEqual(environ['wsgi.url_scheme'], 'http')
+
+
 class TestErrorTask(unittest.TestCase):
 
     def _makeOne(self, channel=None, request=None):
@@ -847,21 +1450,10 @@ class DummyError(object):
 
 class DummyTask(object):
     serviced = False
-    deferred = False
     cancelled = False
 
-    def __init__(self, toraise=None):
-        self.toraise = toraise
-
     def service(self):
         self.serviced = True
-        if self.toraise:
-            raise self.toraise
-
-    def defer(self):
-        self.deferred = True
-        if self.toraise:
-            raise self.toraise
 
     def cancel(self):
         self.cancelled = True
@@ -873,6 +1465,10 @@ class DummyAdj(object):
     port = 80
     url_prefix = ''
     trusted_proxy = None
+    trusted_proxy_count = 1
+    trusted_proxy_headers = set()
+    log_untrusted_proxy_headers = True
+    clear_untrusted_proxy_headers = True
 
 class DummyServer(object):
     server_name = 'localhost'
@@ -885,7 +1481,7 @@ class DummyChannel(object):
     closed_when_done = False
     adj = DummyAdj()
     creation_time = 0
-    addr = ['127.0.0.1']
+    addr = ('127.0.0.1', 39830)
 
     def __init__(self, server=None):
         if server is None:
@@ -924,8 +1520,8 @@ class DummyLogger(object):
     def __init__(self):
         self.logged = []
 
-    def warning(self, msg):
-        self.logged.append(msg)
+    def warning(self, msg, *args):
+        self.logged.append(msg % args)
 
-    def exception(self, msg):
-        self.logged.append(msg)
+    def exception(self, msg, *args):
+        self.logged.append(msg % args)
